#include <sys/types.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

#include "renju.h"
#include "sendmail.h"
#include "strutl.h"


int Renju::IsSymmetric(int row1, int col1, int row2, int col2)
{
	int drow1[4], dcol1[4];
	int drow2[4], dcol2[4];

	// First build array of distances
	int row, col;
	char *comma;
	int m;
	drow1[0] = row1-7, dcol1[0] = col1-7;
	drow2[0] = row2-7, dcol2[0] = col2-7;

	DecodeMove(moves[0], row, col);
	drow1[1] = row1-row, dcol1[1] = col1-col;
	drow2[1] = row2-row, dcol2[1] = col2-col;

	comma = strchr(moves[0],',');
	DecodeMove(comma+1, row, col);
	drow1[2] = row1-row, dcol1[2] = col1-col;
	drow2[2] = row2-row, dcol2[2] = col2-col;

	DecodeMove(moves[1], row, col);
	drow1[3] = row1-row, dcol1[3] = col1-col;
	drow2[3] = row2-row, dcol2[3] = col2-col;

	// Now check if the distances are the same (mirror and rotate)
	int mir,rot;
	for (mir = 0; mir < 2; mir++) {
		for (rot = 0; rot < 4; rot++) {
			// index 0 and 2 are both relative to X and index 1 and 3 to O.
			if (((drow1[0]==drow2[0] && dcol1[0]==dcol2[0] &&
			      drow1[2]==drow2[2] && dcol1[2]==dcol2[2]) ||
			     (drow1[0]==drow2[2] && dcol1[0]==dcol2[2] &&
			      drow1[2]==drow2[0] && dcol1[2]==dcol2[0])) &&
			    ((drow1[1]==drow2[1] && dcol1[1]==dcol2[1] &&
			      drow1[3]==drow2[3] && dcol1[3]==dcol2[3]) ||
			     (drow1[1]==drow2[3] && dcol1[1]==dcol2[3] &&
			      drow1[3]==drow2[1] && dcol1[3]==dcol2[1])))
				return 1;
			for (m = 0; m < 4; m++) {
				int temp = drow2[m];
				drow2[m] = dcol2[m];
				dcol2[m] = -temp;
			}
		}
		for (m = 0; m < 4; m++)
			drow2[m] = -drow2[m];
	}
	return 0; // Not symmetrical
}


const char *Renju::MakeMove(const char *move)
{
	int row, col;
	static char newmove[8];
	char *comma;
	const char *m;
	char mov[100];
	int ix;
	
	switch (moves.Count()) {
	case 0:
			strcpy(mov,move);
			comma = strchr(mov,',');
			if (!comma)
				return (char*)Error("Move #1 must be two stones (white,black)");
			*comma = '\0';
			m = DecodeMove(mov,row,col);
			if (m == NULL) return NULL;
			if (row < 6 || row > 8 || col < 6 || col > 8)
				return (char*)Error("First stone must be [G-J][7-9]");
			if (!IsBlank(row,col))
				return (char*)Error("%s is not vacant", m);
			PutAt(row,col,PlayerPieces(CurrentPlayer()+1));
			last.Init();
			last.Add(row).Add(col);
			strcat(newmove,m);
			strcat(newmove,",");
			m = DecodeMove(comma+1,row,col);
			if (m == NULL) return NULL;
			if (row < 5 || row > 9 || col < 5 || col > 9)
				return (char*)Error("Second stone must be [F-K][6-10]");
			if (!IsBlank(row,col))
				return (char*)Error("%s is not vacant", m);
			PutAt(row,col,PlayerPieces(CurrentPlayer()));
			last.Add(row).Add(col);
			strcat(newmove,m);
			break;
	case 2:
			strcpy(mov,move);
			comma = strchr(mov,',');
			if (!comma)
				return (char*)Error("Move #3 must be two moves");
			*comma = '\0';
			m = DecodeMove(mov,row,col);
			if (m == NULL) return NULL;
			if (!IsBlank(row,col))
				return (char*)Error("%s is not vacant", m);
			PutAt(row,col,'?');
			last.Init();
			last.Add(row).Add(col);
			int row2, col2;
			choice.Init();
			choice.Add(m);
			strcpy(newmove,m);
			strcat(newmove,",");
			m = DecodeMove(comma+1,row2,col2);
			if (m == NULL) return NULL;
			if (!IsBlank(row2,col2))
				return (char*)Error("%s is not vacant", m);
			PutAt(row2,col2,'?');
			last.Add(row2).Add(col2);
			choice.Add(m);
			strcat(newmove,m);
			if (IsSymmetric(row,col,row2,col2))
				return (char*)Error("The move you made results in a symmetrical (illegal) board");
			break;
	case 3:
			strcpy(mov,move);
			comma = strchr(mov,',');
			if (!comma)
				return (char*)Error("Move #4 must be two moves");
			*comma = '\0';
			m = DecodeMove(mov,row,col);
			if (m == NULL) return NULL;
			strcpy(newmove,m);
			strcat(newmove,",");
			ix = choice.Has(m);
			if (ix < 0)
				return (char*)Error("please specify which of %s and %s to keep", choice[0], choice[1]);
			PutAt(row,col,PlayerPieces(CurrentPlayer(-1)));
			DecodeMove(choice[1-ix],row,col);
			PutAt(row,col,Blank());
			m = DecodeMove(comma+1,row,col);
			if (m == NULL) return NULL;
			strcat(newmove,m);
			if (!IsBlank(row,col))
				return (char*)Error("%s is not vacant", m);
			PutAt(row,col,PlayerPieces(CurrentPlayer()));
			last.Init();
			last.Add(row).Add(col);
			choice.Init();
			break;
	default:
			m = DecodeMove(move,row,col);
			if (m == NULL) return NULL;
			strcpy(newmove,m);

			if (!IsBlank(row,col))
				return (char*)Error("%s is not vacant", newmove);
			PutAt(row,col,PlayerPieces(CurrentPlayer()));

			last.Init();
			last.Add(row).Add(col);
			break;
	}
	return newmove;
}

int Renju::MustSkip(void)
{
/* 	if (moves.Count() == 1 || moves.Count() == 3) */
/* 		return 1; */
	return 0;
}

const char *Renju::ForcedMove(void)
{
/* 	if (moves.Count() == 0) */
/* 		return "h8"; */
	return NULL;
}

int Renju::RowDelta[] = {  0,  1,  1,  1 };
int Renju::ColDelta[] = {  1,  0,  1, -1 };

void Renju::Collect56(int row,int col,int d,int q[],int *found,int level)
{
        (void)level;
    
	char color = GetAt(row,col);

	if (color == Blank())
		return;

/* printf("%*sCollect56(%c%d,%c)\n",level*3,"",toupper(Cols[col-MinCol()]),row-MinRow()+1,"-|/\\"[d]); */

	int length=1;
	int j;

	int sign;
	for (sign=-1 ; sign<2 ; sign += 2) {
		for (j=1 ; ; j++) {
			if (GetAt(row+sign*j*RowDelta[d],col+sign*j*ColDelta[d]) != color)
				break;
			length++;
		}
	}
	if (length > 5)
		q[6]++,	(*found)++;
	else if (length == 5)
		q[5]++,	(*found)++;

/* 	if (*found) */
/* 		if (length > 5) */
/* 			printf("%*soverline\n",level*3,""); */
/* 		else */
/* 			printf("%*sfound\n",level*3,""); */
}

void Renju::Collect4(int row,int col,int d,int q[],int *found,int level)
{
        (void)level;
        
	char color = GetAt(row,col);

	if (color != 'x')
		return;

/* printf("%*sCollect4(%c%d,%c)\n",level*3,"",toupper(Cols[col-MinCol()]),row-MinRow()+1,"-|/\\"[d]); */
	int i,j,nwhite,nblack;

	for (i=-4 ; i<=0 ; i++) {
		if (!InsideBoard(row+i*RowDelta[d],col+i*ColDelta[d]) ||
			!InsideBoard(row+(i+4)*RowDelta[d],col+(i+4)*ColDelta[d]))
			continue;
		if (GetAt(row+(i-1)*RowDelta[d],col+(i-1)*ColDelta[d]) == 'x' ||
			GetAt(row+(i+5)*RowDelta[d],col+(i+5)*ColDelta[d]) == 'x')
			continue;
		nwhite=nblack=0;

		for (j=0 ; j<5 ; j++) {
			char piece = GetAt(row+(i+j)*RowDelta[d],col+(i+j)*ColDelta[d]);
			if (piece == 'x')
				nblack++;
			else if (piece == 'o') {
				nwhite++;
				break;
			}
		}

		if (!nwhite && nblack == 4) {
			q[4]++,	(*found)++;

			if (GetAt(row+i*RowDelta[d],col+i*ColDelta[d]) == Blank())
				i++;
		}
	}

/* 	if (*found) */
/* 		printf("%*sfound\n",level*3,""); */
}

void Renju::Collect3(int row,int col,int d,int q[],int *found,int level)
{
/* printf("%*sCollect3(%c%d,%c)\n",level*3,"",toupper(Cols[col-MinCol()]),row-MinRow()+1,"-|/\\"[d]); */
	int i,j,nwhite,nblack;
	int focus;

	for (i=-3 ; i<=0 ; i++) {
		if (!InsideBoard(row+(i-1)*RowDelta[d],col+(i-1)*ColDelta[d]) ||
			!InsideBoard(row+(i+4)*RowDelta[d],col+(i+4)*ColDelta[d]))
			continue;
		if (GetAt(row+(i-1)*RowDelta[d],col+(i-1)*ColDelta[d]) != Blank() ||
			GetAt(row+(i+4)*RowDelta[d],col+(i+4)*ColDelta[d]) != Blank())
			continue;
		if (GetAt(row+(i-2)*RowDelta[d],col+(i-2)*ColDelta[d]) == 'x' ||
			GetAt(row+(i+5)*RowDelta[d],col+(i+5)*ColDelta[d]) == 'x')
			continue;
		nwhite=nblack=0;

		for (j=0 ; j<4 ; j++) {
			char piece = GetAt(row+(i+j)*RowDelta[d],col+(i+j)*ColDelta[d]);
			if (piece == 'x')
				nblack++;
			else if (piece == 'o') {
				nwhite++;
				break;
			}
			else
				focus = i+j;
		}

		if (nwhite || nblack != 3)
			continue;

		if (CheckMove(row+focus*RowDelta[d],col+focus*ColDelta[d],0,level+1) == LEGAL) {
			q[3]++,	(*found)++;
/* printf("%*sfound\n",level*3,""); */
			return;
		}
	}

}

void Renju::CollectRows(int row,int col,int d,int q[],int level)
{
	int found=0;

/* printf("%*sCollectRows(%c%d,%c)\n",level*3,"",toupper(Cols[col-MinCol()]),row-MinRow()+1,"-|\\/"[d]); */

	Collect56(row,col,d,q,&found,level+1);

	if (found || GetAt(row,col) == 'o')
		return;

	Collect4(row,col,d,q,&found,level+1);

	if (found)
		return;

	Collect3(row,col,d,q,&found,level+1);

	return;
}

Renju::MoveType Renju::CheckMove(int row, int col, int Whosturn,int level)
{
	int quantity[7];

/* printf("%*sCheckMove(%c%d,%c)\n",level*3,"",toupper(Cols[col-MinCol()]),row-MinRow()+1,PlayerPieces(Whosturn%2)); */

	if (GetAt(row,col) != Blank())
		return ILLEGAL;

	int i;
	for (i=0 ; i<7 ; i++)
		quantity[i] = 0;

	PutAt(row, col, PlayerPieces(Whosturn%2));

	for (i=0 ; i<4 ; i++)
		CollectRows(row,col,i,quantity,level+1);

	PutAt(row, col, Blank());

	if (Whosturn == 1)	// White
		if (quantity[5] || quantity[6])
			return WINNING;
		else
			return LEGAL;
	else // Black
		if (quantity[5])
			return WINNING;
		else if (quantity[6] || quantity[3] >= 2 || quantity[4] >= 2)
			return FORBIDDEN;
		else
			return LEGAL;
}

int Renju::IsGameOver(const char *&winner)
{
	int row = last[0];
	int col = last[1];
	/* Check for win... */

	char piece = GetAt(row,col);
	PutAt(row,col,Blank());
	MoveType rc = CheckMove(row,col,CurrentPlayer(-1));
	PutAt(row,col,piece);

	switch (rc) {
	case WINNING:
		winner = players[CurrentPlayer(-1)];
		return 1;
	case FORBIDDEN:
		winner = players[CurrentPlayer()];
		return 1;
	default:
		if (!isdigit(boardno[0]) && isdigit(boardno[strlen(boardno)-1])) {
			if (moves.Count() >= 148) {
				winner = NULL;
				return 1;
			}
		}
		return 0;
	}
}

int Renju::ReadBoard(GameFile &game)
{
	int result = Board2D::ReadBoard(game);
	result |= choice.Read(game);
	return result;
}

int Renju::WriteBoard(GameFile &game)
{
	int result = Board2D::WriteBoard(game);
	result *= choice.Write(game);
	return result;

}

int Renju::Init(void)
{
	if (parameters.Count() > 0)
		return Error("%s does not take parameters", GameType());
	choice.Init();
	Gomoku::Init();
	PutAt(7,7,PlayerPieces(0));
	return 1;
}

int Renju::NagTime(void)
{
	if (!isdigit(boardno[0]) && isdigit(boardno[strlen(boardno)-1]))
		return 7;
	return Game::NagTime();
}

int Renju::ForfeitTime(void)
{
	if (!isdigit(boardno[0]) && isdigit(boardno[strlen(boardno)-1]))
		return 21;
	return Game::ForfeitTime();
}
int Renju::TotalTimeAllowed(void)
{
	if (!isdigit(boardno[0]) && isdigit(boardno[strlen(boardno)-1]))
		return 135;
	return Game::TotalTimeAllowed();
}
